/**
 * Visits a TypeScript file and:
 * 1. Gathers all Schema definitions
 * 2. Gather all SQL invocations
 * 4. Processes schemas
 * 5. Processes SQL invocations
 *
 * If a SQL invocation refers to a not yet defined schema, this
 * visitor:
 * 1. Finds the file that declares the schema
 * 2. Visits that file
 * 4. Adds the schema definition to the schema cache
 * 5. Returns to this visitor
 *
 * At which point this visitor will add the current
 * files as depending on the other file.
 *
 * This dependency tracking will allow the `analyzer` to
 * analyze all files that depend on another file
 * when that base file changes.
 *
 * Even if the schema has already been defined, we should still set up a dependency
 * between the file containing `sql` statements and all `schema`s it uses.
 *
 * Dependencies on self are ignored.
 */
import SchemaCache from "../SchemaCache.js";
import ts from "typescript";
import DependencyGraph from "../DependencyGraph.js";
import { Fix } from "./types.js";
import { Options } from "../Analyzer.js";
/**
 * The file visitor is ephemeral. Created each time we visit a file.
 */
export default class FileVisitor {
    #private;
    private options;
    private schemaCache;
    private dag;
    private sourceFile;
    private sqlTemplates;
    private schemaTemplates;
    constructor(options: Options, schemaCache: SchemaCache, dag: DependencyGraph, sourceFile: ts.SourceFile);
    visitAll(checker: ts.TypeChecker): Promise<void>;
    visitSchemaDefs(checker: ts.TypeChecker): Promise<void>;
    visitQueryDefs(checker: ts.TypeChecker): Promise<void>;
    applyFixes(fixes: Fix[]): Promise<void>;
    collectAllNodes(node: ts.Node, checker: ts.TypeChecker): void;
    collectSchemaNodes(node: ts.Node, checker: ts.TypeChecker): void;
    collectQueryNodes(node: ts.Node, checker: ts.TypeChecker): void;
}
//# sourceMappingURL=FileVisitor.d.ts.map