import { getQueryRelations, parseQueryRelations, } from "@vlcn.io/type-gen-ts-adapter";
import { getChildren, normalize, trimTag } from "../util.js";
import ts from "typescript";
export default class QueryTypeBuilder {
    schemaCache;
    dag;
    sourceFile;
    constructor(schemaCache, dag, sourceFile) {
        this.schemaCache = schemaCache;
        this.dag = dag;
        this.sourceFile = sourceFile;
    }
    buildQueryTypes(queries, checker) {
        const fixes = [];
        for (const query of queries) {
            const fix = this.processSqlTemplate(query, checker);
            if (fix != null) {
                fixes.push(fix);
            }
        }
        return fixes;
    }
    processSqlTemplate(node, checker) {
        const children = getChildren(node);
        const templateStringNode = children[children.length - 1];
        const schemaAccessNode = children[0];
        const schemaNode = getChildren(schemaAccessNode)[0];
        // range of text to replace. Inclusive of `<` and `>` if they exist.
        const range = [
            schemaAccessNode.getEnd(),
            templateStringNode.getStart(),
        ];
        const maybeExistingNode = children[1];
        if (ts.isTemplateLiteral(templateStringNode)) {
            try {
                const schemaRelations = this.lookupRelations(schemaNode, checker);
                if (schemaRelations == null) {
                    return null;
                }
                // process it, extracting type information
                let existingContent = "";
                if (maybeExistingNode != templateStringNode) {
                    existingContent = normalize(`<${maybeExistingNode.getText()}>`);
                }
                const replacement = this.genQueryShape(templateStringNode.getText(), schemaRelations);
                if (existingContent == normalize(replacement)) {
                    return null;
                }
                // const pos = this.sourceFile.getLineAndCharacterOfPosition(range[0]);
                return { _tag: "InlineFix", range, replacement };
            }
            catch (e) {
                const existingContent = normalize(`<${maybeExistingNode.getText()}>`);
                const newContent = normalize(`<{/*${e.message}*/}>`);
                if (existingContent == newContent) {
                    return null;
                }
                return {
                    _tag: "InlineFix",
                    range,
                    replacement: `<{/*${e.message}*/}>`,
                };
            }
        }
        throw new Error(`Got an unexpected kind of node: ${templateStringNode.kind}. Was expecting a sql template string.`);
    }
    genQueryShape(query, schemaRelations) {
        query = trimTag(query);
        // TODO: they could have passed many queries...
        try {
            const shape = parseQueryRelations(getQueryRelations(query, schemaRelations))[0];
            if (shape == null) {
                return `<unknown>`;
            }
            else {
                return `<{
  ${Object.entries(shape)
                    .map(([key, value]) => {
                    return `${key}: ${value}`;
                })
                    .join(";\n  ")}
}>`;
            }
        }
        catch (e) {
            return `<{/*
  ${e.message}
*/}>`;
        }
    }
    /**
     * The ordering in which we process files guarantees that the cache is already warm with
     * schema definitions.
     */
    lookupRelations(schemaNode, checker) {
        // const schemaNodeSymbol = checker.getSymbolAtLocation(schemaNode);
        const type = checker.getTypeAtLocation(schemaNode);
        const prop = type.getProperty("__type");
        // must not be _our_ sql literal but someone else's
        if (prop == null) {
            return null;
        }
        const internalType = checker.getTypeOfSymbol(prop);
        // console.log(internalType.getSymbol()?.declarations);
        const internalProps = checker.getPropertiesOfType(internalType);
        const decl = internalType.getSymbol()?.declarations?.[0];
        if (!decl) {
            const loc = this.sourceFile.getLineAndCharacterOfPosition(schemaNode.getStart());
            throw new Error(`Could not find the referenced schema typescript type! Is it defined? ${this.sourceFile.fileName}:${loc.line}:${loc.character}`);
        }
        // console.log(
        //   "source? ",
        //   decl.getSourceFile().fileName,
        //   decl.getStart(),
        //   decl.getText()
        // );
        if (decl.getSourceFile().fileName != this.sourceFile.fileName) {
            this.dag.addDependent(decl.getSourceFile().fileName, this.sourceFile.fileName);
        }
        const ret = this.schemaCache.get(decl.getSourceFile().fileName, normalize(decl.getFullText()));
        if (!ret) {
            const loc = this.sourceFile.getLineAndCharacterOfPosition(schemaNode.getStart());
            throw new Error(`Could not find the referenced schema relations! Are they defined? ${this.sourceFile.fileName}:${loc.line}:${loc.character}}`);
        }
        return ret;
    }
}
/**
 * Cache relation by
 * type_structure -> relation
 *
 * Still blow away on processing file that defines the type structures.
 *
 * Map dag by:
 * file -> type_structures -> dependent files
 */
//# sourceMappingURL=QueryTypeBuilder.js.map